@@ -30,9 +30,9 @@
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <inttypes.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syslog.h>
 #include <unistd.h>
 
 #include "config.h"
@@ -76,6 +76,8 @@ dhcp_print_option_encoding(const struct dhcp_opt *opt, int cols)
 		printf(" ip6address");
 	else if (opt->type & FLAG)
 		printf(" flag");
+	else if (opt->type & BITFLAG)
+		printf(" bitflags");
 	else if (opt->type & RFC3397)
 		printf(" domain");
 	else if (opt->type & DOMAIN)
@@ -86,9 +88,9 @@ dhcp_print_option_encoding(const struct dhcp_opt *opt, int cols)
 		printf(" raw");
 	else if (opt->type & BINHEX)
 		printf(" binhex");
-	else if (opt->type & STRING)	
+	else if (opt->type & STRING)
 		printf(" string");
-	if (opt->type & RFC3361)	
+	if (opt->type & RFC3361)
 		printf(" rfc3361");
 	if (opt->type & RFC3442)
 		printf(" rfc3442");
@@ -152,9 +154,9 @@ make_option_mask(const struct dhcp_opt *dopts, size_t dopts_len,
     const struct dhcp_opt *odopts, size_t odopts_len,
     uint8_t *mask, const char *opts, int add)
 {
-	char *token, *o, *p, *t;
+	char *token, *o, *p;
 	const struct dhcp_opt *opt;
-	int match;
+	int match, e;
 	unsigned int n;
 	size_t i;
 
@@ -169,11 +171,10 @@ make_option_mask(const struct dhcp_opt *dopts, size_t dopts_len,
 			if (strcmp(opt->var, token) == 0)
 				match = 1;
 			else {
-				errno = 0;
-				n = (unsigned int)strtol(token, &t, 0);
-				if (errno == 0 && !*t)
-					if (opt->option == n)
-						match = 1;
+				n = (unsigned int)strtou(token, NULL, 0,
+				    0, UINT_MAX, &e);
+				if (e == 0 && opt->option == n)
+					match = 1;
 			}
 			if (match)
 				break;
@@ -183,11 +184,10 @@ make_option_mask(const struct dhcp_opt *dopts, size_t dopts_len,
 				if (strcmp(opt->var, token) == 0)
 				        match = 1;
 				else {
-					errno = 0;
-					n = (unsigned int)strtol(token, &t, 0);
-					if (errno == 0 && !*t)
-						if (opt->option == n)
-							match = 1;
+					n = (unsigned int)strtou(token, NULL, 0,
+					    0, UINT_MAX, &e);
+					if (e == 0 && opt->option == n)
+						match = 1;
 				}
 				if (match)
 					break;
@@ -457,14 +457,15 @@ print_string(char *dst, size_t len, int type, const uint8_t *data, size_t dl)
 			errno = EINVAL;
 			break;
 		}
-		if (!(type & (ASCII | RAW | ESCSTRING)) /*plain string */ &&
+		if (!(type & (ASCII | RAW | ESCSTRING | ESCFILE)) /* plain */ &&
 		    (!isascii(c) && !isprint(c)))
 		{
 			errno = EINVAL;
 			break;
 		}
-		if (type & ESCSTRING &&
-		    (c == '\\' || !isascii(c) || !isprint(c)))
+		if ((type & (ESCSTRING | ESCFILE) &&
+		    (c == '\\' || !isascii(c) || !isprint(c))) ||
+		    (type & ESCFILE && (c == '/' || c == ' ')))
 		{
 			errno = EINVAL;
 			if (c == '\\') {
@@ -485,9 +486,9 @@ print_string(char *dst, size_t len, int type, const uint8_t *data, size_t dl)
 					return -1;
 				}
 				*dst++ = '\\';
-		                *dst++ = (((unsigned char)c >> 6) & 03) + '0';
-		                *dst++ = (((unsigned char)c >> 3) & 07) + '0';
-		                *dst++ = ( (unsigned char)c       & 07) + '0';
+		                *dst++ = (char)(((c >> 6) & 03) + '0');
+		                *dst++ = (char)(((c >> 3) & 07) + '0');
+		                *dst++ = (char)(( c       & 07) + '0');
 				len -= 4;
 			}
 			bytes += 4;
@@ -560,7 +561,7 @@ dhcp_optlen(const struct dhcp_opt *opt, size_t dl)
 		sz = sizeof(uint32_t);
 	else if (opt->type & UINT16)
 		sz = sizeof(uint16_t);
-	else if (opt->type & UINT8)
+	else if (opt->type & (UINT8 | BITFLAG))
 		sz = sizeof(uint8_t);
 	else if (opt->type & ADDRIPV6)
 		sz = ADDR6SZ;
@@ -570,15 +571,9 @@ dhcp_optlen(const struct dhcp_opt *opt, size_t dl)
 	return (dl < sz ? 0 : sz);
 }
 
-#ifdef INET6
-#define PO_IFNAME
-#else
-#define PO_IFNAME __unused
-#endif
-
-ssize_t
-print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
-    PO_IFNAME const char *ifname)
+static ssize_t
+print_option(char *s, size_t len, const struct dhcp_opt *opt,
+    const uint8_t *data, size_t dl, const char *ifname)
 {
 	const uint8_t *e, *t;
 	uint16_t u16;
@@ -590,7 +585,11 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 	size_t l;
 	char *tmp;
 
-	if (type & RFC3397) {
+#ifndef INET6
+	UNUSED(ifname);
+#endif
+
+	if (opt->type & RFC3397) {
 		sl = decode_rfc3397(NULL, 0, data, dl);
 		if (sl == 0 || sl == -1)
 			return sl;
@@ -599,32 +598,32 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 		if (tmp == NULL)
 			return -1;
 		decode_rfc3397(tmp, l, data, dl);
-		sl = print_string(s, len, type, (uint8_t *)tmp, l - 1);
+		sl = print_string(s, len, opt->type, (uint8_t *)tmp, l - 1);
 		free(tmp);
 		return sl;
 	}
 
 #ifdef INET
-	if (type & RFC3361) {
+	if (opt->type & RFC3361) {
 		if ((tmp = decode_rfc3361(data, dl)) == NULL)
 			return -1;
 		l = strlen(tmp);
-		sl = print_string(s, len, type, (uint8_t *)tmp, l);
+		sl = print_string(s, len, opt->type, (uint8_t *)tmp, l);
 		free(tmp);
 		return sl;
 	}
 
-	if (type & RFC3442)
+	if (opt->type & RFC3442)
 		return decode_rfc3442(s, len, data, dl);
 
-	if (type & RFC5969)
+	if (opt->type & RFC5969)
 		return decode_rfc5969(s, len, data, dl);
 #endif
 
-	if (type & STRING)
-		return print_string(s, len, type, data, dl);
+	if (opt->type & STRING)
+		return print_string(s, len, opt->type, data, dl);
 
-	if (type & FLAG) {
+	if (opt->type & FLAG) {
 		if (s) {
 			*s++ = '1';
 			*s = '\0';
@@ -632,27 +631,50 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 		return 1;
 	}
 
+	if (opt->type & BITFLAG) {
+		/* bitflags are a string, MSB first, such as ABCDEFGH
+		 * where A is 10000000, B is 01000000, etc. */
+		bytes = 0;
+		for (l = 0, sl = sizeof(opt->bitflags) - 1;
+		    l < sizeof(opt->bitflags);
+		    l++, sl--)
+		{
+			/* Don't print NULL or 0 flags */
+			if (opt->bitflags[l] != '\0' &&
+			    opt->bitflags[l] != '0' &&
+			    *data & (1 << sl))
+			{
+				if (s)
+					*s++ = opt->bitflags[l];
+				bytes++;
+			}
+		}
+		if (s)
+			*s = '\0';
+		return bytes;
+	}
+
 	if (!s) {
-		if (type & UINT8)
+		if (opt->type & UINT8)
 			l = 3;
-		else if (type & UINT16) {
+		else if (opt->type & UINT16) {
 			l = 5;
 			dl /= 2;
-		} else if (type & SINT16) {
+		} else if (opt->type & SINT16) {
 			l = 6;
 			dl /= 2;
-		} else if (type & UINT32) {
+		} else if (opt->type & UINT32) {
 			l = 10;
 			dl /= 4;
-		} else if (type & SINT32) {
+		} else if (opt->type & SINT32) {
 			l = 11;
 			dl /= 4;
-		} else if (type & ADDRIPV4) {
+		} else if (opt->type & ADDRIPV4) {
 			l = 16;
 			dl /= 4;
 		}
 #ifdef INET6
-		else if (type & ADDRIPV6) {
+		else if (opt->type & ADDRIPV6) {
 			e = data + dl;
 			l = 0;
 			while (data < e) {
@@ -681,36 +703,36 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 			bytes++;
 			len--;
 		}
-		if (type & UINT8) {
+		if (opt->type & UINT8) {
 			sl = snprintf(s, len, "%u", *data);
 			data++;
-		} else if (type & UINT16) {
+		} else if (opt->type & UINT16) {
 			memcpy(&u16, data, sizeof(u16));
 			u16 = ntohs(u16);
 			sl = snprintf(s, len, "%u", u16);
 			data += sizeof(u16);
-		} else if (type & SINT16) {
+		} else if (opt->type & SINT16) {
 			memcpy(&u16, data, sizeof(u16));
 			s16 = (int16_t)ntohs(u16);
 			sl = snprintf(s, len, "%d", s16);
 			data += sizeof(u16);
-		} else if (type & UINT32) {
+		} else if (opt->type & UINT32) {
 			memcpy(&u32, data, sizeof(u32));
 			u32 = ntohl(u32);
 			sl = snprintf(s, len, "%u", u32);
 			data += sizeof(u32);
-		} else if (type & SINT32) {
+		} else if (opt->type & SINT32) {
 			memcpy(&u32, data, sizeof(u32));
 			s32 = (int32_t)ntohl(u32);
 			sl = snprintf(s, len, "%d", s32);
 			data += sizeof(u32);
-		} else if (type & ADDRIPV4) {
+		} else if (opt->type & ADDRIPV4) {
 			memcpy(&addr.s_addr, data, sizeof(addr.s_addr));
 			sl = snprintf(s, len, "%s", inet_ntoa(addr));
 			data += sizeof(addr.s_addr);
 		}
 #ifdef INET6
-		else if (type & ADDRIPV6) {
+		else if (opt->type & ADDRIPV6) {
 			ssize_t r;
 
 			r = ipv6_printaddr(s, len, data, ifname);
@@ -731,8 +753,40 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 	return bytes;
 }
 
+int
+dhcp_set_leasefile(char *leasefile, size_t len, int family,
+    const struct interface *ifp, const char *extra)
+{
+	char ssid[len];
+
+	if (ifp->name[0] == '\0') {
+		strlcpy(leasefile, ifp->ctx->pidfile, len);
+		return 0;
+	}
+
+	switch (family) {
+	case AF_INET:
+	case AF_INET6:
+		break;
+	default:
+		errno = EINVAL;
+		return -1;
+	}
+
+	if (ifp->wireless) {
+		ssid[0] = '-';
+		print_string(ssid + 1, sizeof(ssid) - 1,
+		    ESCFILE,
+		    (const uint8_t *)ifp->ssid, ifp->ssid_len);
+	} else
+		ssid[0] = '\0';
+	return snprintf(leasefile, len,
+	    family == AF_INET ? LEASEFILE : LEASEFILE6,
+	    ifp->name, ssid, extra);
+}
+
 static size_t
-dhcp_envoption1(char **env, const char *prefix,
+dhcp_envoption1(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
     const struct dhcp_opt *opt, int vname, const uint8_t *od, size_t ol,
     const char *ifname)
 {
@@ -742,7 +796,7 @@ dhcp_envoption1(char **env, const char *prefix,
 
 	if (opt->len && opt->len < ol)
 		ol = opt->len;
-	len = print_option(NULL, 0, opt->type, od, ol, ifname);
+	len = print_option(NULL, 0, opt, od, ol, ifname);
 	if (len < 0)
 		return 0;
 	if (vname)
@@ -756,7 +810,7 @@ dhcp_envoption1(char **env, const char *prefix,
 		return e;
 	v = val = *env = malloc(e);
 	if (v == NULL) {
-		syslog(LOG_ERR, "%s: %m", __func__);
+		logger(ctx, LOG_ERR, "%s: %m", __func__);
 		return 0;
 	}
 	if (vname)
@@ -764,7 +818,7 @@ dhcp_envoption1(char **env, const char *prefix,
 	else
 		v += snprintf(val, e, "%s=", prefix);
 	if (len != 0)
-		print_option(v, (size_t)len + 1, opt->type, od, ol, ifname);
+		print_option(v, (size_t)len + 1, opt, od, ol, ifname);
 	return e;
 }
 
@@ -785,7 +839,8 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 
 	/* If no embedded or encapsulated options, it's easy */
 	if (opt->embopts_len == 0 && opt->encopts_len == 0) {
-		if (dhcp_envoption1(env == NULL ? NULL : &env[0],
+		if (!(opt->type & RESERVED) &&
+		    dhcp_envoption1(ctx, env == NULL ? NULL : &env[0],
 		    prefix, opt, 1, od, ol, ifname))
 			return 1;
 		return 0;
@@ -796,7 +851,7 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 		if (opt->type & INDEX) {
 			if (opt->index > 999) {
 				errno = ENOBUFS;
-				syslog(LOG_ERR, "%s: %m", __func__);
+				logger(ctx, LOG_ERR, "%s: %m", __func__);
 				return 0;
 			}
 		}
@@ -804,7 +859,7 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 		    (opt->type & INDEX ? 3 : 0);
 		pfx = malloc(e);
 		if (pfx == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
+			logger(ctx, LOG_ERR, "%s: %m", __func__);
 			return 0;
 		}
 		if (opt->type & INDEX)
@@ -826,10 +881,12 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 		/* Use the option prefix if the embedded option
 		 * name is different.
 		 * This avoids new_fqdn_fqdn which would be silly. */
-		ov = strcmp(opt->var, eopt->var);
-		if (dhcp_envoption1(env == NULL ? NULL : &env[n],
-		    pfx, eopt, ov, od, e, ifname))
-			n++;
+		if (!(eopt->type & RESERVED)) {
+			ov = strcmp(opt->var, eopt->var);
+			if (dhcp_envoption1(ctx, env == NULL ? NULL : &env[n],
+			    pfx, eopt, ov, od, e, ifname))
+				n++;
+		}
 		od += e;
 		ol -= e;
 	}
