@@ -32,6 +32,10 @@
 #include <net/if.h>
 
 #include "config.h"
+#ifdef HAVE_SYS_QUEUE_H
+#include <sys/queue.h>
+#endif
+
 #include "defs.h"
 #include "control.h"
 #include "if-options.h"
@@ -60,15 +64,12 @@ struct interface {
 	struct dhcpcd_ctx *ctx;
 	TAILQ_ENTRY(interface) next;
 	char name[IF_NAMESIZE];
-#ifdef __linux
+#ifdef __linux__
 	char alias[IF_NAMESIZE];
 #endif
 	unsigned int index;
 	unsigned int flags;
 	sa_family_t family;
-#ifdef __FreeBSD__
-	struct sockaddr_storage linkaddr;
-#endif
 	unsigned char hwaddr[HWADDR_LEN];
 	uint8_t hwlen;
 	unsigned int metric;
@@ -84,11 +85,12 @@ struct interface {
 TAILQ_HEAD(if_head, interface);
 
 struct dhcpcd_ctx {
-#ifdef USE_SIGNALS
-	sigset_t sigset;
-#endif
+	int pid_fd;
+	char pidfile[sizeof(PIDFILE) + IF_NAMESIZE + 1];
 	const char *cffile;
 	unsigned long long options;
+	char *logfile;
+	int log_fd;
 	int argc;
 	char **argv;
 	int ifac;	/* allowed interfaces */
@@ -101,11 +103,13 @@ struct dhcpcd_ctx {
 	char **ifcv;	/* configured interfaces */
 	unsigned char *duid;
 	size_t duid_len;
-	int pid_fd;
 	int link_fd;
 	struct if_head *ifaces;
 
-	struct eloop_ctx *eloop;
+#ifdef USE_SIGNALS
+	sigset_t sigset;
+#endif
+	struct eloop *eloop;
 
 	int control_fd;
 	int control_unpriv_fd;
@@ -121,6 +125,7 @@ struct dhcpcd_ctx {
 	struct dhcp_opt *dhcp_opts;
 	size_t dhcp_opts_len;
 	struct rt_head *ipv4_routes;
+	struct rt_head *ipv4_kroutes;
 
 	int udp_fd;
 	uint8_t *packet;
@@ -134,6 +139,8 @@ struct dhcpcd_ctx {
 	unsigned char secret[SECRET_LEN];
 	size_t secret_len;
 
+	struct dhcp_opt *nd_opts;
+	size_t nd_opts_len;
 	struct dhcp_opt *dhcp6_opts;
 	size_t dhcp6_opts_len;
 	struct ipv6_ctx *ipv6;
@@ -151,11 +158,11 @@ struct dhcpcd_ctx {
 };
 
 #ifdef USE_SIGNALS
-extern const int dhcpcd_handlesigs[];
+extern const int dhcpcd_signals[];
+extern const size_t dhcpcd_signals_len;
 #endif
 
 int dhcpcd_oneup(struct dhcpcd_ctx *);
-int dhcpcd_ipwaited(struct dhcpcd_ctx *);
 pid_t dhcpcd_daemonise(struct dhcpcd_ctx *);
 
 int dhcpcd_handleargs(struct dhcpcd_ctx *, struct fd_list *, int, char **);
@@ -167,6 +174,6 @@ void dhcpcd_dropinterface(struct interface *, const char *);
 int dhcpcd_selectprofile(struct interface *, const char *);
 
 void dhcpcd_startinterface(void *);
-void dhcpcd_initstate(struct interface *);
+void dhcpcd_initstate(struct interface *, unsigned long long);
 
 #endif
