@@ -326,7 +326,7 @@ get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
 	size_t buflen;
 	struct nlmsghdr *nlm;
 	struct sockaddr_nl nladdr;
-	socklen_t nladdr_len = sizeof(nladdr);
+	socklen_t nladdr_len;
 	int r;
 
 	buflen = 0;
@@ -351,6 +351,8 @@ get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
 				goto eexit;
 			buf = nbuf;
 		}
+		nladdr.nl_pid = 0;
+		nladdr_len = sizeof(nladdr);
 		bytes = recvfrom(fd, buf, buflen, flags,
 		    (struct sockaddr *)&nladdr, &nladdr_len);
 		if (bytes == -1 || bytes == 0)
@@ -408,6 +410,11 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 		return -1;
 
 	memset(rt, 0, sizeof(*rt));
+	if (rtm->rtm_type == RTN_UNREACHABLE)
+		rt->flags = RTF_REJECT;
+	if (rtm->rtm_scope == RT_SCOPE_HOST)
+		rt->flags |= RTF_HOST;
+
 	prefsrc.s_addr = INADDR_ANY;
 	rta = (struct rtattr *)RTM_RTA(rtm);
 	len = RTM_PAYLOAD(nlm);
@@ -426,7 +433,7 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 			    sizeof(prefsrc.s_addr));
 			break;
 		case RTA_OIF:
-			rt->iface = if_findindex(ctx,
+			rt->iface = if_findindex(ctx->ifaces,
 			    *(unsigned int *)RTA_DATA(rta));
 			break;
 		case RTA_PRIORITY:
@@ -468,9 +475,11 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 		return -1;
 
 	memset(rt, 0, sizeof(*rt));
-	ipv6_mask(&rt->net, rtm->rtm_dst_len);
 	if (rtm->rtm_type == RTN_UNREACHABLE)
 		rt->flags = RTF_REJECT;
+	if (rtm->rtm_scope == RT_SCOPE_HOST)
+		rt->flags |= RTF_HOST;
+	ipv6_mask(&rt->net, rtm->rtm_dst_len);
 
 	rta = (struct rtattr *)RTM_RTA(rtm);
 	len = RTM_PAYLOAD(nlm);
@@ -485,7 +494,7 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 			    sizeof(rt->gate.s6_addr));
 			break;
 		case RTA_OIF:
-			rt->iface = if_findindex(ctx,
+			rt->iface = if_findindex(ctx->ifaces,
 			    *(unsigned int *)RTA_DATA(rta));
 			break;
 		case RTA_PRIORITY:
@@ -589,7 +598,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		return -1;
 	}
 	ifa = NLMSG_DATA(nlm);
-	if ((ifp = if_findindex(ctx, ifa->ifa_index)) == NULL) {
+	if ((ifp = if_findindex(ctx->ifaces, ifa->ifa_index)) == NULL) {
 		/* We don't know about the interface the address is for
 		 * so it's not really an error */
 		return 1;
@@ -618,7 +627,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 			rta = RTA_NEXT(rta, len);
 		}
 		ipv4_handleifa(ctx, nlm->nlmsg_type, NULL, ifp->name,
-		    &addr, &net, &dest);
+		    &addr, &net, &dest, ifa->ifa_flags);
 		break;
 #endif
 #ifdef INET6
@@ -796,8 +805,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 		return 0;
 
 	/* Check for a new interface */
-	ifp = if_find(ctx, ifn);
-	if (ifp == NULL) {
+	if ((ifp = if_find(ctx->ifaces, ifn)) == NULL) {
 		/* If are listening to a dev manager, let that announce
 		 * the interface rather than the kernel. */
 		if (dev_listening(ctx) < 1)
@@ -869,7 +877,7 @@ static int
 add_attr_l(struct nlmsghdr *n, unsigned short maxlen, unsigned short type,
     const void *data, unsigned short alen)
 {
-	unsigned short len = RTA_LENGTH(alen);
+	unsigned short len = (unsigned short)RTA_LENGTH(alen);
 	struct rtattr *rta;
 
 	if (NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len) > maxlen) {
@@ -918,7 +926,7 @@ nla_next(struct nlattr *nla, size_t *rem)
 }
 
 #define NLA_TYPE(nla) ((nla)->nla_type & NLA_TYPE_MASK)
-#define NLA_LEN(nla) ((nla)->nla_len - NLA_HDRLEN)
+#define NLA_LEN(nla) (unsigned int)((nla)->nla_len - NLA_HDRLEN)
 #define NLA_OK(nla, rem) \
 	((rem) >= sizeof(struct nlattr) && \
 	(nla)->nla_len >= sizeof(struct nlattr) && \
@@ -1118,7 +1126,7 @@ struct nlmr
 const char *if_pfname = "Packet Socket";
 
 int
-if_openrawsocket(struct interface *ifp, int protocol)
+if_openrawsocket(struct interface *ifp, uint16_t protocol)
 {
 	int s;
 	union sockunion {
@@ -1186,7 +1194,7 @@ eexit:
 }
 
 ssize_t
-if_sendrawpacket(const struct interface *ifp, int protocol,
+if_sendrawpacket(const struct interface *ifp, uint16_t protocol,
     const void *data, size_t len)
 {
 	const struct dhcp_state *state;
@@ -1218,7 +1226,7 @@ if_sendrawpacket(const struct interface *ifp, int protocol,
 }
 
 ssize_t
-if_readrawpacket(struct interface *ifp, int protocol,
+if_readrawpacket(struct interface *ifp, uint16_t protocol,
     void *data, size_t len, int *flags)
 {
 	struct iovec iov = {
@@ -1359,10 +1367,9 @@ if_route(unsigned char cmd, const struct rt *rt)
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_PREFSRC,
 		    &state->addr.s_addr, sizeof(state->addr.s_addr));
 	}
-	/* If destination == gateway then don't add the gateway */
+	/* If a host route then don't add the gateway */
 	if ((cmd == RTM_ADD || cmd == RTM_CHANGE) &&
-	    (rt->dest.s_addr != rt->gate.s_addr ||
-	    rt->net.s_addr != INADDR_BROADCAST))
+	    rt->net.s_addr != INADDR_BROADCAST)
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_GATEWAY,
 		    &rt->gate.s_addr, sizeof(rt->gate.s_addr));
 
@@ -1407,6 +1414,15 @@ if_initrt(struct interface *ifp)
 	return send_netlink(ifp->ctx, ifp,
 	    NETLINK_ROUTE, &nlm.hdr, &_if_initrt);
 }
+
+int
+if_addrflags(__unused const struct in_addr *addr,
+    __unused const struct interface *ifp)
+{
+
+	/* Linux has no support for IPv4 address flags */
+	return 0;
+}
 #endif
 
 #ifdef INET6
