@@ -28,6 +28,7 @@
 #define _WITH_GETLINE /* Stop FreeBSD bitching */
 
 #include <sys/param.h>
+#include <sys/stat.h>
 #include <sys/types.h>
 
 #include <arpa/inet.h>
@@ -97,6 +98,9 @@
 #define O_IPV6RA_AUTOCONF	O_BASE + 38
 #define O_IPV6RA_NOAUTOCONF	O_BASE + 39
 #define O_REJECT		O_BASE + 40
+#define O_IPV6RA_ACCEPT_NOPUBLIC	O_BASE + 41
+#define O_BOOTP			O_BASE + 42
+#define O_DEFINEND		O_BASE + 43
 
 const struct option cf_options[] = {
 	{"background",      no_argument,       NULL, 'b'},
@@ -156,6 +160,7 @@ const struct option cf_options[] = {
 	{"ipv6ra_fork",     no_argument,       NULL, O_IPV6RA_FORK},
 	{"ipv6ra_own",      no_argument,       NULL, O_IPV6RA_OWN},
 	{"ipv6ra_own_default", no_argument,    NULL, O_IPV6RA_OWN_D},
+	{"ipv6ra_accept_nopublic", no_argument, NULL, O_IPV6RA_ACCEPT_NOPUBLIC},
 	{"ipv4only",        no_argument,       NULL, '4'},
 	{"ipv6only",        no_argument,       NULL, '6'},
 	{"ipv4",            no_argument,       NULL, O_IPV4},
@@ -171,6 +176,7 @@ const struct option cf_options[] = {
 	{"dev",             required_argument, NULL, O_DEV},
 	{"nodev",           no_argument,       NULL, O_NODEV},
 	{"define",          required_argument, NULL, O_DEFINE},
+	{"definend",        required_argument, NULL, O_DEFINEND},
 	{"define6",         required_argument, NULL, O_DEFINE6},
 	{"embed",           required_argument, NULL, O_EMBED},
 	{"encap",           required_argument, NULL, O_ENCAP},
@@ -188,6 +194,7 @@ const struct option cf_options[] = {
 	{"gateway",         no_argument,       NULL, O_GATEWAY},
 	{"ia_pd_mix",       no_argument,       NULL, O_PFXDLGMIX},
 	{"reject",          required_argument, NULL, O_REJECT},
+	{"bootp",           no_argument,       NULL, O_BOOTP},
 	{NULL,              0,                 NULL, '\0'}
 };
 
@@ -514,11 +521,22 @@ set_option_space(struct dhcpcd_ctx *ctx,
 {
 
 #if !defined(INET) && !defined(INET6)
-	/* Satisfy use */
-	ctx = NULL;
+	UNUSED(ctx);
 #endif
 
 #ifdef INET6
+	if (strncmp(arg, "nd_", strlen("nd_")) == 0) {
+		*d = ctx->nd_opts;
+		*dl = ctx->nd_opts_len;
+		*od = ifo->nd_override;
+		*odl = ifo->nd_override_len;
+		*request = ifo->requestmasknd;
+		*require = ifo->requiremasknd;
+		*no = ifo->nomasknd;
+		*reject = ifo->rejectmasknd;
+		return arg + strlen("nd_");
+	}
+
 	if (strncmp(arg, "dhcp6_", strlen("dhcp6_")) == 0) {
 		*d = ctx->dhcp6_opts;
 		*dl = ctx->dhcp6_opts_len;
@@ -628,7 +646,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 	int e, i, t;
 	long l;
 	unsigned long u;
-	char *p = NULL, *fp, *np, **nconf;
+	char *p = NULL, *bp, *fp, *np, **nconf;
 	ssize_t s;
 	struct in_addr addr, addr2;
 	in_addr_t *naddr;
@@ -1234,6 +1252,9 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 	case O_IPV6RA_OWN_D:
 		ifo->options |= DHCPCD_IPV6RA_OWN_DEFAULT;
 		break;
+	case O_IPV6RA_ACCEPT_NOPUBLIC:
+		ifo->options |= DHCPCD_IPV6RA_ACCEPT_NOPUBLIC;
+		break;
 	case O_IPV6RA_AUTOCONF:
 		ifo->options |= DHCPCD_IPV6RA_AUTOCONF;
 		break;
@@ -1458,6 +1479,12 @@ err_sla:
 		dop = &ifo->dhcp_override;
 		dop_len = &ifo->dhcp_override_len;
 		/* FALLTHROUGH */
+	case O_DEFINEND:
+		if (dop == NULL) {
+			dop = &ifo->nd_override;
+			dop_len = &ifo->nd_override_len;
+		}
+		/* FALLTHROUGH */
 	case O_DEFINE6:
 		if (dop == NULL) {
 			dop = &ifo->dhcp6_override;
@@ -1527,13 +1554,18 @@ err_sla:
 		/* length */
 		if (np) {
 			*np++ = '\0';
+			bp = NULL; /* No bitflag */
 			l = (long)strtou(np, NULL, 0, 0, LONG_MAX, &e);
 			if (e) {
-				logger(ctx, LOG_ERR, "failed to convert length");
+				logger(ctx,LOG_ERR, "failed to convert length");
 				return -1;
 			}
-		} else
+		} else {
 			l = 0;
+			bp = strchr(arg, '='); /* bitflag assignment */
+			if (bp)
+				*bp++ = '\0';
+		}
 		t = 0;
 		if (strcasecmp(arg, "request") == 0) {
 			t |= REQUEST;
@@ -1582,6 +1614,8 @@ err_sla:
 			t |= STRING;
 		else if (strcasecmp(arg, "byte") == 0)
 			t |= UINT8;
+		else if (strcasecmp(arg, "bitflags") == 0)
+			t |= BITFLAG;
 		else if (strcasecmp(arg, "uint16") == 0)
 			t |= UINT16;
 		else if (strcasecmp(arg, "int16") == 0)
@@ -1629,6 +1663,11 @@ err_sla:
 			logger(ctx, LOG_WARNING, "ignoring array for strings");
 			t &= ~ARRAY;
 		}
+		if (t & BITFLAG) {
+			if (bp == NULL)
+				logger(ctx, LOG_WARNING,
+				    "missing bitflag assignment");
+		}
 		/* variable */
 		if (!fp) {
 			if (!(t & OPTION)) {
@@ -1642,10 +1681,16 @@ err_sla:
 			fp = strwhite(arg);
 			if (fp)
 				*fp++ = '\0';
-			np = strdup(arg);
-			if (np == NULL) {
-				logger(ctx, LOG_ERR, "%s: %m", __func__);
-				return -1;
+			if (strcasecmp(arg, "reserved")) {
+				np = strdup(arg);
+				if (np == NULL) {
+					logger(ctx, LOG_ERR,
+					    "%s: %m", __func__);
+					return -1;
+				}
+			} else {
+				np = NULL;
+				t |= RESERVED;
 			}
 		}
 		if (opt != O_EMBED) {
@@ -1680,11 +1725,25 @@ err_sla:
 		ndop->type = t;
 		ndop->len = (size_t)l;
 		ndop->var = np;
+		if (bp) {
+			dl = strlen(bp);
+			memcpy(ndop->bitflags, bp, dl);
+			memset(ndop->bitflags + dl, 0,
+			    sizeof(ndop->bitflags) - dl);
+		} else
+			memset(ndop->bitflags, 0, sizeof(ndop->bitflags));
 		/* Save the define for embed and encap options */
-		if (opt == O_DEFINE || opt == O_DEFINE6 || opt == O_VENDOPT)
+		switch (opt) {
+		case O_DEFINE:
+		case O_DEFINEND:
+		case O_DEFINE6:
+		case O_VENDOPT:
 			*ldop = ndop;
-		else if (opt == O_ENCAP)
+			break;
+		case O_ENCAP:
 			*edop = ndop;
+			break;
+		}
 		break;
 	case O_VENDCLASS:
 		fp = strwhite(arg);
@@ -1952,6 +2011,9 @@ err_sla:
 	case O_PFXDLGMIX:
 		ifo->options |= DHCPCD_PFXDLGMIX;
 		break;
+	case O_BOOTP:
+		ifo->options |= DHCPCD_BOOTP;
+		break;
 	default:
 		return 0;
 	}
@@ -2030,6 +2092,7 @@ read_config(struct dhcpcd_ctx *ctx,
 {
 	struct if_options *ifo;
 	FILE *fp;
+	struct stat sb;
 	char *line, *buf, *option, *p;
 	size_t buflen;
 	ssize_t vlen;
@@ -2088,6 +2151,14 @@ read_config(struct dhcpcd_ctx *ctx,
 			ifo->dhcp_override_len = INITDEFINES;
 #endif
 
+#if defined(INET6) && defined(INITDEFINENDS)
+		ifo->nd_override =
+		    calloc(INITDEFINENDS, sizeof(*ifo->nd_override));
+		if (ifo->nd_override == NULL)
+			logger(ctx, LOG_ERR, "%s: %m", __func__);
+		else
+			ifo->nd_override_len = INITDEFINENDS;
+#endif
 #if defined(INET6) && defined(INITDEFINE6S)
 		ifo->dhcp6_override =
 		    calloc(INITDEFINE6S, sizeof(*ifo->dhcp6_override));
@@ -2160,15 +2231,24 @@ read_config(struct dhcpcd_ctx *ctx,
 		ifo->dhcp_override_len = 0;
 
 #ifdef INET6
+		ctx->nd_opts = ifo->nd_override;
+		ctx->nd_opts_len = ifo->nd_override_len;
 		ctx->dhcp6_opts = ifo->dhcp6_override;
 		ctx->dhcp6_opts_len = ifo->dhcp6_override_len;
 #else
+		for (i = 0, opt = ifo->nd_override;
+		    i < ifo->nd_override_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->nd_override);
 		for (i = 0, opt = ifo->dhcp6_override;
 		    i < ifo->dhcp6_override_len;
 		    i++, opt++)
 			free_dhcp_opt_embenc(opt);
 		free(ifo->dhcp6_override);
 #endif
+		ifo->nd_override = NULL;
+		ifo->nd_override_len = 0;
 		ifo->dhcp6_override = NULL;
 		ifo->dhcp6_override_len = 0;
 
@@ -2186,6 +2266,8 @@ read_config(struct dhcpcd_ctx *ctx,
 		free(buf);
 		return ifo;
 	}
+	if (stat(ctx->cffile, &sb) == 0)
+		ifo->mtime = sb.st_mtime;
 
 	ldop = edop = NULL;
 	while ((line = get_line(&buf, &buflen, fp))) {
@@ -2320,6 +2402,11 @@ free_options(struct if_options *ifo)
 		    opt++, ifo->dhcp_override_len--)
 			free_dhcp_opt_embenc(opt);
 		free(ifo->dhcp_override);
+		for (opt = ifo->nd_override;
+		    ifo->nd_override_len > 0;
+		    opt++, ifo->nd_override_len--)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->nd_override);
 		for (opt = ifo->dhcp6_override;
 		    ifo->dhcp6_override_len > 0;
 		    opt++, ifo->dhcp6_override_len--)
