@@ -30,6 +30,11 @@
 
 #include "dhcpcd.h"
 
+#ifdef IN_IFF_TENTATIVE
+#define IN_IFF_NOTUSEABLE \
+        (IN_IFF_TENTATIVE | IN_IFF_DUPLICATED | IN_IFF_DETACHED)
+#endif
+
 struct rt {
 	TAILQ_ENTRY(rt) next;
 	struct in_addr dest;
@@ -40,7 +45,7 @@ struct rt {
 	unsigned int metric;
 #endif
 	struct in_addr src;
-	uint8_t flags;
+	unsigned int flags;
 };
 TAILQ_HEAD(rt_head, rt);
 
@@ -50,6 +55,7 @@ struct ipv4_addr {
 	struct in_addr net;
 	struct in_addr dst;
 	struct interface *iface;
+	int addr_flags;
 };
 TAILQ_HEAD(ipv4_addrhead, ipv4_addr);
 
@@ -69,12 +75,18 @@ int ipv4_ifcmp(const struct interface *, const struct interface *);
 uint8_t inet_ntocidr(struct in_addr);
 int inet_cidrtoaddr(int, struct in_addr *);
 uint32_t ipv4_getnetmask(uint32_t);
+int ipv4_ifaddrexists(const struct interface *);
 int ipv4_addrexists(struct dhcpcd_ctx *, const struct in_addr *);
 
 #define STATE_ADDED		0x01
 #define STATE_FAKE		0x02
+#define STATE_TENTATIVE		0x04
 
 void ipv4_buildroutes(struct dhcpcd_ctx *);
+void ipv4_finaliseaddr(struct interface *);
+int ipv4_deladdr(struct interface *ifp, const struct in_addr *,
+    const struct in_addr *);
+int ipv4_preferanother(struct interface *ifp);
 void ipv4_applyaddr(void *);
 int ipv4_handlert(struct dhcpcd_ctx *, int, struct rt *);
 void ipv4_freerts(struct rt_head *);
@@ -84,7 +96,8 @@ struct ipv4_addr *ipv4_iffindaddr(struct interface *,
 struct ipv4_addr *ipv4_iffindlladdr(struct interface *);
 struct ipv4_addr *ipv4_findaddr(struct dhcpcd_ctx *, const struct in_addr *);
 void ipv4_handleifa(struct dhcpcd_ctx *, int, struct if_head *, const char *,
-    const struct in_addr *, const struct in_addr *, const struct in_addr *);
+    const struct in_addr *, const struct in_addr *, const struct in_addr *,
+    int);
 
 void ipv4_freeroutes(struct rt_head *);
 
@@ -97,7 +110,7 @@ void ipv4_ctxfree(struct dhcpcd_ctx *);
 #define ipv4_freeroutes(a) {}
 #define ipv4_free(a) {}
 #define ipv4_ctxfree(a) {}
-#define ipv4_addrexists(a, b) (0)
+#define ipv4_ifaddrexists(a) (0)
 #endif
 
 #endif
