@@ -86,21 +86,18 @@ if_printoptions(void)
 		printf(" -  %s\n", *p);
 }
 
-#ifdef USE_SIGNALS
-#define U
-#else
-#define U __unused
-#endif
 static int
-exec_script(U const struct dhcpcd_ctx *ctx, char *const *argv, char *const *env)
-#undef U
+exec_script(const struct dhcpcd_ctx *ctx, char *const *argv, char *const *env)
 {
 	pid_t pid;
 	posix_spawnattr_t attr;
-	int i;
+	int r;
 #ifdef USE_SIGNALS
+	size_t i;
 	short flags;
 	sigset_t defsigs;
+#else
+	UNUSED(ctx);
 #endif
 
 	/* posix_spawn is a safe way of executing another image
@@ -111,15 +108,15 @@ exec_script(U const struct dhcpcd_ctx *ctx, char *const *argv, char *const *env)
 	flags = POSIX_SPAWN_SETSIGMASK | POSIX_SPAWN_SETSIGDEF;
 	posix_spawnattr_setflags(&attr, flags);
 	sigemptyset(&defsigs);
-	for (i = 0; dhcpcd_handlesigs[i]; i++)
-		sigaddset(&defsigs, dhcpcd_handlesigs[i]);
+	for (i = 0; i < dhcpcd_signals_len; i++)
+		sigaddset(&defsigs, dhcpcd_signals[i]);
 	posix_spawnattr_setsigdefault(&attr, &defsigs);
 	posix_spawnattr_setsigmask(&attr, &ctx->sigset);
 #endif
 	errno = 0;
-	i = posix_spawn(&pid, argv[0], NULL, &attr, argv, env);
-	if (i) {
-		errno = i;
+	r = posix_spawn(&pid, argv[0], NULL, &attr, argv, env);
+	if (r) {
+		errno = r;
 		return -1;
 	}
 	return pid;
@@ -288,7 +285,7 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 	if (ifp->ctx->options & DHCPCD_DUMPLEASE)
 		elen = 2;
 	else
-		elen = 13;
+		elen = 12;
 
 #define EMALLOC(i, l) if ((env[(i)] = malloc((l))) == NULL) goto eexit;
 	/* Make our env + space for profile, wireless and debug */
@@ -375,12 +372,6 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 		env[11] = strdup("if_oneup=false");
 	if (env[11] == NULL)
 		goto eexit;
-	if (dhcpcd_ipwaited(ifp->ctx))
-		env[12] = strdup("if_ipwaited=true");
-	else
-		env[12] = strdup("if_ipwaited=false");
-	if (env[12] == NULL)
-		goto eexit;
 	if (ifo->options & DHCPCD_DEBUG) {
 		e = strlen("syslog_debug=true") + 1;
 		EMALLOC(elen, e);
